;; This program computes and prints a table of joypad directional button
;; combinations and their effect on the cursor in the name entry menu. The
;; computed table is the source of data for game.NAME-ENTRY-CURSOR-MOVEMENT.

(local fennel (require :fennel))

;; Unload our own modules, to avoid old versions being cached in package.loaded
;; from a previous execution.
(each [_ modname (ipairs [:ff1 :ff1na :game :savestate :ucs :util])]
  (tset package.loaded modname nil))
;; Here we load a version-specific ff1 module, then additionally store it as
;; package.loaded.ff1. Other modules will require it simply as :ff1.
(local ff1 (require :ff1na)) (tset package.loaded :ff1 ff1)
(local game (require :game))
(local
  {: joypad-set
   : joypad-press-release
   : read-u8
   : next-frame
   : await-exec
   : earliest-effective-frame
   : run}
  (require :savestate))
(local ucs (require :ucs))

(lambda earliest-effective-frame-0 [...] (let [x (select 1 (earliest-effective-frame ...))] x))
(lambda earliest-effective-frame-1 [...] (let [x (select 2 (earliest-effective-frame ...))] x))
(lambda earliest-effective-frame-2 [...] (let [x (select 3 (earliest-effective-frame ...))] x))

(local {: Up : Down : Left : Right : Start : Select : B : A} game)

(lambda party-select-class [savestate]
  (earliest-effective-frame-1
    savestate
    #(joypad-press-release $ A)
    #(await-exec $ 8 ff1.CODE.DoNameInput)))

;; An iterator over the powerset (the set of all subsets) of the elements in
;; the sequential table elems, where a set is represented by a table mapping
;; elements to booleans. Each iteratior returns an index for for the subset
;; (from 0 to 2^n − 1, where n is the length of elems), as well as a table
;; representing the subset.
(lambda powerset [elems]
  (lambda f [elems n]
    (let [n (+ n 1)]
      ;; Interpret n as a binary number, where a 1 bit meaning to include the
      ;; corresponding element, and a 0 bit meaning to exclude it.
      (if (< n (lshift 1 (length elems)))
        (values n
                (collect [i e (ipairs elems)]
                  (let [mask (lshift 1 (- i 1))]
                    (values e (not= 0 (band n mask)))))))))
  (values f elems -1))

;; A canonical ordering of joypad button names.
(local JOYPAD-BUTTONS [Up Down Left Right Start Select B A])

;; Convert a joypad button→boolean table into an integer bitmap.
(lambda joypad->bitmap [joypad]
  (accumulate [bitmap 0 i button (ipairs JOYPAD-BUTTONS)]
    (let [mask (lshift 1 (- i 1))]
      (bor bitmap (if (. joypad button) mask 0)))))

;; Convert an integer bitmap into a joypad button→boolean table.
(lambda bitmap->joypad [bitmap]
  (collect [i button (ipairs JOYPAD-BUTTONS)]
    (let [mask (lshift 1 (- i 1))]
      (values button (not= 0 (band bitmap mask))))))

;; Compute a 16×16 table: starting with a subset of the buttons [Up Down Left
;; Right] pressed, and then changing to a different subset of those same
;; buttons, what is the effect on the game's [namecurs_x namecurs_y] variables?
;; At the top level, the table keys are joypad bitmaps; below that are tables
;; that map joypad bitmaps to differences in the three variables.
;;
;; The assumption here is that the relative effect of joypad inputs is the
;; same, no matter the starting position of the game's cursor variables.
;; Therefore, this table (derived empirically through emulation) can be used
;; to predict how the cursor variables will change in reaction to new joypad
;; inputs, from any initial configuration.
(lambda compute-name-entry-transitions [savestate]
  (lambda mod-interval [n low high]
    (+ low (% (- n low) (- high low))))
  (lambda mod-nearest [n width]
    (let [low (- (// width 2))
          high (+ low width)]
      (mod-interval n low high)))
  (collect [_ from-joypad (powerset [Up Down Left Right])]
    (let [from-savestate (-> savestate
                             (joypad-set from-joypad)
                             (next-frame)
                             (next-frame))
          from-x (read-u8 from-savestate ff1.VAR.namecurs_x)
          from-y (read-u8 from-savestate ff1.VAR.namecurs_y)]
      (values
        (joypad->bitmap from-joypad)
        (collect [_ to-joypad (powerset [Up Down Left Right])]
          (let [to-savestate (-> from-savestate
                                 (joypad-set to-joypad)
                                 (next-frame)
                                 (next-frame))
                to-x (read-u8 to-savestate ff1.VAR.namecurs_x)
                to-y (read-u8 to-savestate ff1.VAR.namecurs_y)]
            (values
              (joypad->bitmap to-joypad)
              {:namecurs_x (mod-nearest (- to-x from-x) ff1.NAME-ENTRY-WIDTH)
               :namecurs_y (mod-nearest (- to-y from-y) ff1.NAME-ENTRY-HEIGHT)})))))))

(lambda main [savestate]
  (-> savestate
      (game.title-menu)
      (party-select-class)
      (compute-name-entry-transitions)
      (fennel.view)
      (print)))

(assert (run main))
