;; These are subroutines for high-level game actions, such as moving on the map
;; and navigating menus.

(local fennel (require :fennel))

(local ff1 (require :ff1)) ; Must already have been loaded into package.loaded.
(local
  {: joypad-set
   : joypad-press-release
   : joypad-release-all
   : read-u8
   : next-frame
   : await-exec
   : await-u8-satisfies
   : await-u8-changes
   : await-frame-satisfies
   : earliest-effective-frame}
  (require :savestate))
(local ucs (require :ucs))
(local util (require :util))

(local Up     "P1 Up")
(local Down   "P1 Down")
(local Left   "P1 Left")
(local Right  "P1 Right")
(local Start  "P1 Start")
(local Select "P1 Select")
(local B      "P1 B")
(local A      "P1 A")
(local Reset  "Reset")

(lambda earliest-effective-frame-0 [...] (let [x (select 1 (earliest-effective-frame ...))] x))
(lambda earliest-effective-frame-1 [...] (let [x (select 2 (earliest-effective-frame ...))] x))
(lambda earliest-effective-frame-2 [...] (let [x (select 3 (earliest-effective-frame ...))] x))

(lambda set-respond-rate [savestate respond-rate]
  (let [current (read-u8 savestate ff1.VAR.respondrate)
        left-steps  (% (- current respond-rate) 8)
        right-steps (% (- respond-rate current) 8)
        ;; Go left or right, whichever is closer.
        (button steps sign) (if (< left-steps right-steps)
                              (values Left  left-steps  -1)
                              (values Right right-steps +1))]
    (faccumulate [s savestate i 1 steps]
      (earliest-effective-frame-1
        s
        #(joypad-press-release $ button)
        #(await-u8-satisfies $ 4 ff1.VAR.respondrate #(= $ (% (+ current (* sign i)) 8)))))))

(lambda title-menu [savestate respond-rate]
  (-> savestate
      ;; Press the appropriate button to advance past the story. JP uses Reset
      ;; and NA uses Start.
      (earliest-effective-frame-1
        #(joypad-press-release $ (case ff1.VERSION :jp Reset :na Start _ (error)))
        #(await-exec $ 8 ff1.CODE.EnterTitleScreen))
      (set-respond-rate respond-rate)
      ;; Select "Continue" to start a new game. "Continue" is the same as "New
      ;; Game" if there is no saved game yet.
      ;; In the NA version we can use A or Start, but Start is better because
      ;; it saves 1 frame. In the JP version we can only use A.
      (earliest-effective-frame-1
        #(joypad-press-release $ (case ff1.VERSION :jp A :na Start _ (error)))
        #(-?> $
              (await-u8-satisfies 4 ff1.VAR.respondrate #(= $ respond-rate))
              (await-exec 8 ff1.CODE.DoPartyGen_OnCharacter)))))

(lambda party-select-class [savestate n class]
  (let [ptygen (. ff1.VAR.ptygen n)]
    (lambda scroll-to [savestate class]
      ;; Any directional button works to advance through the cycle of character
      ;; classes. Alternating two different buttons saves a frame over always
      ;; using the same button.
      (accumulate [s savestate _ button (util.irepeat [Left Right])
                   &until (await-u8-satisfies s 4 ptygen.class #(= $ class))]
        (let [current-class (read-u8 s ptygen.class)]
          ;; Press the button and see if ptygen.class changes.
          (earliest-effective-frame-2 s
                                      #(joypad-press-release $ button)
                                      #(await-u8-changes $ 4 ptygen.class)))))
    (-> savestate
        ;; Press buttons until the right class is selected.
        (scroll-to class)
        ;; Press A to confirm.
        (earliest-effective-frame-1
          #(joypad-press-release $ A)
          #(await-exec $ 8 ff1.CODE.DoNameInput)))))

;; This table stores the effect on the position of the name entry cursor for
;; selected combinations of directional buttons. These are stored in the order
;; in which party-enter-name will try them. There is only one combination that
;; moves up, and only one combination that moves left; but there are multiple
;; ways to move down or right. The table includes two combinations for moving
;; down and right; this gives the optimizer in party-enter-name the option of
;; alternating between them for every-frame movement when moving in one of
;; those directions.
;;
;; An aspect not reflected in this table is that the cursor moves only if the
;; directional inputs are different from the previous frame. That part of the
;; logic is implemented in party-enter-name.
;;
;; https://tasvideos.org/GameResources/NES/FinalFantasy1#NameEntry
;;
;; The program compute-name-entry-transitions.fnl computes the full 16×16 table
;; of joypad directional button combinations and their effect on the cursor.
(local NAME-ENTRY-CURSOR-MOVEMENT
  [{:joypad {Up false Down false Left false Right false} :diff {:x  0 :y  0}} ; .... → no change
   {:joypad {Up  true Down false Left false Right false} :diff {:x  0 :y -1}} ; U... → up
   {:joypad {Up false Down  true Left false Right false} :diff {:x  0 :y +1}} ; .D.. → down
   {:joypad {Up false Down false Left  true Right false} :diff {:x -1 :y  0}} ; ..L. → left
   {:joypad {Up false Down false Left false Right  true} :diff {:x +1 :y  0}} ; ...R → right
   {:joypad {Up false Down  true Left  true Right false} :diff {:x  0 :y +1}} ; .DL. → down
   {:joypad {Up false Down false Left  true Right  true} :diff {:x +1 :y  0}} ; ..LR → right
   ;; All other combinations would yield {:x 0 :y +1} (down).
   ])

;; Here, name is a sequential table of up to 4 character codes, as in
;; ff1.CHARACTER-ENCODING.
(lambda party-enter-name [savestate n name]
  ;; Callbacks for uniform-cost-search.
  (lambda key [state]
    ;; We have a distinguished "initial" state because we want to be able to
    ;; try all joypad combinations as a first step. If the start state had no
    ;; buttons pressed and no flag to distinguish itself as somehow
    ;; different, it would prohibit transitions back to a no-buttons-pressed
    ;; state.
    (string.format "%d,%d,%d,%d,%s,%s,%s,%s,%s"
                   (if state.initial 1 0)
                   state.cursor
                   state.namecurs_x state.namecurs_y
                   (if (. state.joypad Up)    "U" "_")
                   (if (. state.joypad Down)  "D" "_")
                   (if (. state.joypad Left)  "L" "_")
                   (if (. state.joypad Right) "R" "_")
                   (if (. state.joypad A)     "A" "_")))
  (lambda finished? [state]
    (= 4 state.cursor))
  (lambda neighbors [state]
    (let [selectedtile (. ff1.CONST.lut_NameInput (+ (* 10 state.namecurs_y) state.namecurs_x 1))
          ;; Are we pointing at the next letter in the name? (Or, if past the
          ;; end of the name, at a space character 0xff?)
          on-target? (= selectedtile (or (. name (+ state.cursor 1)) 0xff))
          neighbors
          (if (and on-target? (. state.joypad A))
            ;; If we're over the next letter in name and pressing the A button,
            ;; we need to release the A button (we don't get to combine the
            ;; release for free with a cursor movement).
            [{:cursor state.cursor
              :namecurs_x state.namecurs_x
              :namecurs_y state.namecurs_y
              :joypad (util.table-update state.joypad {A false})}]
            on-target?
            ;; If we're over the next letter in name and not pressing A, then
            ;; press A to advance cursor.
            [{:cursor (+ state.cursor 1)
              :namecurs_x state.namecurs_x
              :namecurs_y state.namecurs_y
              :joypad (util.table-update state.joypad {A true})}]
            ;; Otherwise, we need to move to the target letter. Try pressing
            ;; directional buttons.
            (icollect [_ {: joypad : diff} (ipairs NAME-ENTRY-CURSOR-MOVEMENT)]
              ;; The cursor doesn't move unless at least one directional button
              ;; changes. Ignore the button combination that doesn't work.
              (if (not (and (= (. joypad Up)    (. state.joypad Up))
                            (= (. joypad Down)  (. state.joypad Down))
                            (= (. joypad Left)  (. state.joypad Left))
                            (= (. joypad Right) (. state.joypad Right))))
                {:cursor state.cursor
                 :namecurs_x (% (+ state.namecurs_x diff.x) ff1.NAME-ENTRY-WIDTH)
                 :namecurs_y (% (+ state.namecurs_y diff.y) ff1.NAME-ENTRY-HEIGHT)
                 :joypad (util.table-update joypad {A false})})))]
      ;; Return an iterator through the neighbors table.
      (var i 0)
      (lambda []
        (set i (+ i 1))
        ;; Report a fixed cost of 1 frame for every state transition, even
        ;; though pressing A takes longer than pressing a directional button.
        ;; There are always the same number of A presses per name entry, so
        ;; it doesn't affect optimization.
        (values (. neighbors i) 1))))

  (lambda optimize [savestate n name]
    (let [ptygen (. ff1.VAR.ptygen n)
          start {:initial true
                 :cursor     (read-u8 savestate ff1.VAR.cursor)
                 :namecurs_x (read-u8 savestate ff1.VAR.namecurs_x)
                 :namecurs_y (read-u8 savestate ff1.VAR.namecurs_y)
                 :joypad {}}
          path (assert (ucs.uniform-cost-search start key finished? neighbors)
                       (string.format "no way to enter name %s" (fennel.view name)))]
      ;; Trace through the path to arrive at a finished savestate.
      (accumulate [s savestate
                   i {: cursor : namecurs_x : namecurs_y : joypad}
                   (ipairs path)]
        (earliest-effective-frame-1
          s
          #(joypad-set $ joypad)
          #(await-frame-satisfies $ 4 #(and
                                         ;; Wait until the cursors are pointing where we expect.
                                         (= cursor     (read-u8 $ ff1.VAR.cursor))
                                         (= namecurs_x (read-u8 $ ff1.VAR.namecurs_x))
                                         (= namecurs_y (read-u8 $ ff1.VAR.namecurs_y))
                                         ;; And the most recent character (if any) is correct.
                                         (or (= cursor 0) (= (read-u8 $ (+ ptygen.name (- cursor 1))) (or (. name cursor) 0xff)))))))))

  (-> savestate
      ;; Wait for the name entry cursor variables to be initialized.
      (earliest-effective-frame-2
        #$
        #(await-frame-satisfies $ 4 #(and
                                       (= 0 (read-u8 $ ff1.VAR.cursor))
                                       (= 0 (read-u8 $ ff1.VAR.namecurs_x))
                                       (= 0 (read-u8 $ ff1.VAR.namecurs_y)))))
      ;; Run the shortest-path optimization.
      (optimize n name)
      (joypad-release-all 1)
      ;; Press A once more to confirm the name.
      (earliest-effective-frame-1
        #(joypad-press-release $ A)
        #(await-exec $ 8 ff1.CODE.PtyGen_DrawScreen))))

(lambda party-select [savestate c1 c2 c3 c4]
  (let [savestate (accumulate [s savestate
                               i {: class : name}
                               (ipairs [c1 c2 c3 c4])]
                    (-> s
                        (party-select-class i class)
                        (party-enter-name i name)))
        ;; Final confirmation of character class and name entry.
        savestate (earliest-effective-frame-1
                    savestate
                    #(joypad-press-release $ A)
                    #(await-exec $ 32 ff1.CODE.DoOverworld))]
    savestate))

;; Press the given direction button at the first opportunity to move. The
;; returned savestate is at the frame where the button is pressed.
(lambda move-on-map [savestate direction]
  (let [
        ;; In order to avoid losing time, we need to enter the input for the
        ;; next step 1 frame before the previous frame finishes (before
        ;; move_speed and move_ctr_x/move_ctr_y become zero). There may already
        ;; be an input entered on this frame to take a step, in which case we
        ;; want to let this step happen before entering the input for the next
        ;; step. Peek 1 frame ahead to see if move_speed becomes nonzero. If it
        ;; does not, go ahead and enter the input on this frame. If it does,
        ;; wait until move_speed is 1 frame from becoming zero, and then enter
        ;; the input.
        savestate
        (if (= 0 (read-u8 (next-frame savestate) ff1.VAR.move_speed))
          savestate
          (earliest-effective-frame-1
            (next-frame savestate)
            #$
            #(await-u8-satisfies $ 1 ff1.VAR.move_speed #(= $ 0))))]
    ;; We now know we're at the end of a previous step. Now find the earliest
    ;; frame at which pressing the button affects move_speed. (This is
    ;; immediate if we're already in map movement, but may not be immediate if
    ;; we're just transitioning to the overworld, for example.)
    (earliest-effective-frame-1
      savestate
      #(joypad-press-release $ direction)
      #(await-u8-satisfies $ 2 ff1.VAR.move_speed #(not= $ 0)))))

(lambda await-screen-wipe [savestate]
  (assert (await-exec savestate 128 ff1.CODE.ScreenWipe_Finalize)))

(lambda await-screen-wipe-2 [savestate]
  (-> savestate
      (await-screen-wipe)
      (await-screen-wipe)))

(lambda await-palette-cycle [savestate]
  (assert (await-exec savestate 128 ff1.CODE.CyclePalettes_Done)))

(lambda open-menu [savestate]
  (earliest-effective-frame-1
    savestate
    #(joypad-press-release $ Start)
    #(await-exec $ 128 ff1.CODE.EnterMainMenu)))

(lambda close-menu [savestate]
  (earliest-effective-frame-1
    savestate
    #(joypad-press-release $ B)
    #(await-exec $ 24 ff1.CODE.MainMenuLoop_B_Pressed)))

{: Up
 : Down
 : Left
 : Right
 : Start
 : Select
 : B
 : A

 : title-menu
 : party-select
 : move-on-map
 : await-screen-wipe
 : await-screen-wipe-2
 : await-palette-cycle
 : open-menu
 : close-menu}
