(local bizhawk (require :bizhawk))
(local util (require :util))

;; The functions whose names are suffixed with `!` are "raw", in that they
;; depend on or modify the live emulator state. In contrast, the savestate
;; objects exported by this module are immutable and do not depend on the live
;; emulator state.

;; Save the current state of the emulator to an on-disk file. Returns true on
;; success, or nil and an error message on failure.
(lambda save-to-file! [path ?suppressosd]
  (let [suppressosd (if (= nil ?suppressosd) false ?suppressosd)]
    ;; In BizHawk 2.10 and earlier, savestate.save has no return value; instead
    ;; it signals failure by raising an error. After
    ;; https://github.com/TASEmulators/BizHawk/commit/f409cc445d19f52c0d6da05ec921d67280d7334f
    ;; (BizHawk 2.10.1), savestate.save returns a boolean that indicates
    ;; success status. We use pcall to be compatible with the older
    ;; convention--but if pcall does actually catch an error, the program will
    ;; crash anyway: https://tasvideos.org/Forum/Posts/534691.
    (case (pcall bizhawk.savestate.save path suppressosd)
      (false ?err) (values nil ?err)
      (true false) (values nil (string.format "failed to save savestate to file %q" path))
      _ true)))

;; A "native savestate" is a BizHawk-native savestate, low-level and identified
;; by an ID. A native savestate is one component of what in this module is
;; called a "savestate" (which also contains other information, such as a
;; joypad state).

;; Define SAVESTATE-DIRECTORY to a path (e.g. "/tmp") in order to save
;; temporary savestates to disk rather than use in-memory core savestates. If
;; you want to save the results of program execution as a movie file, you have
;; to use actual on-disk savestates, not in-memory core savestates. This is
;; because the latter do not store input logs.
(local SAVESTATE-DIRECTORY nil)
(local
  {: save-native-savestate!
   : load-native-savestate!
   : delete-native-savestate!}
  (case SAVESTATE-DIRECTORY
    ;; If SAVESTATE-DIRECTORY is not set, use memorysavestate savestates.
    nil {:save-native-savestate! (lambda [] (bizhawk.memorysavestate.savecorestate))
         :load-native-savestate! (lambda [id] (bizhawk.memorysavestate.loadcorestate id))
         :delete-native-savestate! (lambda [id] (bizhawk.memorysavestate.removestate id))}
    ;; Otherwise, use disk file savestates.
    dir (let [basename (table.concat (fcollect [_ 1 16]
                                       (let [r (math.random 16)]
                                         (string.sub "0123456789abcdef" r r))))]
          ;; Generate temporary savestate filenames from a random basename and
          ;; an incrementing counter.
          (var counter 0)
          (lambda generate-path []
            (set counter (+ counter 1))
            (string.format "%s/tempsavestate-%s-%010d" dir basename counter))
          {:save-native-savestate! (lambda []
                                     (let [path (generate-path)]
                                       (assert (save-to-file! path true))
                                       path))
           :load-native-savestate! (lambda [id]
                                     (assert (bizhawk.savestate.load id true)))
           :delete-native-savestate! (lambda [id]
                                       (assert (os.remove id)))})))

;; Check that id has the expected format for an event listener ID. Return the
;; ID if so; otherwise raise an error.
;;
;; When a request to register an event listener fails internally, BizHawk may
;; return an "empty" GUID string "00000000-0000-0000-0000-000000000000" rather
;; than return nil or raise an error. For example,
;; https://github.com/TASEmulators/BizHawk/blob/2.10/src/BizHawk.Client.Common/lua/LuaHelperLibs/EventsLuaLibrary.cs#L136
;;     return Guid.Empty.ToString();
(lambda assert-event-id [id]
  (assert (= (type id) "string")
          (string.format "purported event id %s has type %q, not %q"
                         id (type id) "string"))
  (assert (string.match id "^%x%x%x%x%x%x%x%x%-%x%x%x%x%-%x%x%x%x%-%x%x%x%x%-%x%x%x%x%x%x%x%x%x%x%x%x$")
          (string.format "purported event id %q is not a GUID" id))
  (assert (not= id "00000000-0000-0000-0000-000000000000")
          (string.format "event id %q is all zero" id))
  id)

;; Run the emulator until the next onframestart event. The resulting aligned
;; emulator state is safe to make or load a native savestate from. See
;; discussion starting at
;; https://github.com/TASEmulators/BizHawk/issues/1430#issuecomment-1987446287.
(lambda frame-align! []
  (let [co (coroutine.running)
        joypad (bizhawk.joypad.get)]
    (var id nil)
    (set id (assert-event-id
              (bizhawk.event.onframestart (lambda []
                                            (assert (bizhawk.event.unregisterbyid id))
                                            (assert (coroutine.resume co))))))
    ;; Yield, and wait for onframestart to wake us back up.
    (coroutine.yield)
    ;; Restore the joypad buttons from before the alignment.
    (bizhawk.joypad.set joypad)))

;; Within a savestate object, we wrap a native statestate ID in a table and set
;; a metatable to call delete-native-savestate! on it when it is longer
;; referenced. This lets us share multiple references to the same native
;; savestate from different savestate objects.
(local NATIVE-SAVESTATE-MT
  {:__gc #(delete-native-savestate! $.id)})

;; Create a new savestate object from the current state of the emulator. It is
;; the caller's responsibility to ensure that the emulator state is
;; frame-aligned.
;;
;; A savestate object consists of a native savestate ID, the state of joypad
;; buttons, and a schedule of joypad changes to take effect in future frames.
(lambda new-savestate! []
  {:native (setmetatable {:id (save-native-savestate!)}
                         NATIVE-SAVESTATE-MT)
   :joypad (bizhawk.joypad.get)
   :joypad-schedule []})

;; Make the live emulator state match the given savestate object.
;;
;; This function is used internally by the savestate module for low-level
;; emulator control. In a high-level script, you probably only want to call it
;; once, at the end, in order to save a movie that ends at the final computed
;; savestate.
(lambda restore-savestate! [savestate]
  ;; Native savestates do not contain the joypad state. Set the emulator joypad
  ;; state from the savestate object.
  (bizhawk.joypad.set savestate.joypad)
  ;; The emulator must be aligned at a frame boundary before loading a
  ;; savestate. The emulator state is otherwise unimportant, because we are
  ;; about to throw it away.
  (frame-align!)
  ;; Load the native savestate.
  (load-native-savestate! savestate.native.id))

;; Return a new savestate with a scheduled joypad changes. buttons is a table
;; that maps button names to Boolean values. Buttons that are not mentioned do
;; not change. ?time is how many frames in the future the joypad change should
;; take effect. If ?time is 0 or nil, it affects the joypad state in the
;; current frame.
(lambda joypad-set [savestate buttons ?time]
  (each [k (pairs buttons)]
    (assert (not= nil (. savestate.joypad k))
            (string.format "not a known joypad button: %q" k)))
  (case (or ?time 0)
    ;; If time is 0, directly update the current buttons table.
    0 (util.table-update savestate
                         {:joypad (util.table-update savestate.joypad buttons)})
    ;; If time is in the future, add the buttons to the schedule.
    (where time (> time 0))
    (util.table-update savestate
                       {:joypad-schedule (util.table-append savestate.joypad-schedule
                                                            {:time time
                                                             :buttons (util.table-update buttons)})})
    ?time (error (string.format "invalid time offset: %q" ?time))))

;; Return a new savestate with a scheduled joypad button press followed by a
;; release. button is a button name. ?time is how many frames in the future the
;; press should occur. If ?time is 0 or nil, the press occurs in the current
;; frame. ?duration is how many frames after the press occurs the release
;; should occur. The default for ?duration is 1.
(lambda joypad-press-release [savestate button ?time ?duration]
  (let [time (or ?time 0)
        duration (or ?duration 1)]
    (-> savestate
        (joypad-set {button true} time)
        (joypad-set {button false} (+ time duration)))))

;; Return a new savestate with a scheduled release of all joypad buttons. ?time
;; is how many frames in the future the releases should occur. If ?time is 0 or
;; nil, the releases occur in the current frame.
(lambda joypad-release-all [savestate ?time]
  (joypad-set savestate
              (collect [k _ (pairs savestate.joypad)] (values k false))
              ?time))

;; The joypad-schedule in a savestate object is a sequential table of
;; {:time :buttons} tables. The order of entries in the sequental table doesn't
;; matter, except for the relative positions of entries with equal :time. This
;; function returns a table of button state changes that should happen on the
;; next frame, and a new joypad-schedule that is advanced by one frame. It
;; works by decrementing the :time of all entries in the table by 1, removing
;; any entries that become 0 and accumuating their :buttons into a table of
;; button state changes.
(lambda joypad-schedule-advance [joypad-schedule]
  (let [new-buttons {}
        new-joypad-schedule []]
    (each [_ {: time : buttons} (ipairs joypad-schedule)]
      (case time
        ;; time 1 becomes time 0. Omit from new-joypad-schedule and accumulate
        ;; buttons into new-buttons.
        1 (collect [k v (pairs buttons) &into new-buttons] (values k v))
        ;; Decrement time by 1 and include in new-joypad-schedule.
        time (table.insert new-joypad-schedule {:time (- time 1) : buttons})))
    (values new-buttons new-joypad-schedule)))

;; Return the result of running f, after setting the live emulator state to
;; match savestate.
(lambda with-savestate! [savestate f]
  (restore-savestate! savestate)
  (f))

;; Return the frame count of savestate.
(lambda frame-count [savestate]
  (with-savestate! savestate #(bizhawk.emu.framecount)))

;; Return the value of the u8 at address addr in savestate.
(lambda read-u8 [savestate addr]
  (with-savestate! savestate #(bizhawk.memory.read_u8 addr)))

;; Return a new savestate that represents the state of the emulator after the
;; first of any number of events occurs; or nil if a frame limit is exceeded.
;; The ... arguments are event registration functions that return an event
;; listener ID, such as bizhawk.event.on_bus_exec or bizhawk.event.onframeend.
;; When one of the events occurs, return a new savestate and the index of event
;; registration function. If ?frame-limit is not nil, it is a limit on how many
;; future frames to try. If ?frame-limit frames elapse without any event
;; occurring, this function returns nil.
(lambda await-event [savestate ?frame-limit ...]
  (if (and ?frame-limit (<= ?frame-limit 0))
    nil
    (do
      (var ?co (coroutine.running))
      ;; This result closure is called when one of the registered events happens,
      ;; or when ?frame-limit expires. In the former case, ?event-index is the
      ;; index of the event that happened first. In the latter case, ?event-index
      ;; is nil.
      (lambda result [?event-index]
        (case ?co
          co (do
               ;; We want coroutine.resume to be called only once. Simply
               ;; unregistering all of the event listeners at this point would not
               ;; guarantee that other event listeners would not also call
               ;; coroutine.resume. This is because more than one listener may be
               ;; registered for the same event. (In particular, this happens when
               ;; one of the user's events is bizhawk.event.onframeend, which is
               ;; the same event used for counting up to ?frame-limit.) BizHawk
               ;; runs all callbacks that were registered at the time the event
               ;; occurred, even if, in the course of working through the list of
               ;; callbacks, one of them becomes unregistered. We set ?co to nil to
               ;; signal to future invocations of this function that they should
               ;; not call coroutine.resume again.
               (set ?co nil)
               (assert (coroutine.resume co ?event-index)))))
      ;; Reset the emulator state to the given savestate. We are going to register
      ;; a bunch of event listeners, then wait for something to happen and the
      ;; emulator to yield control back to us. Then we will unregister all the
      ;; listeners.
      (restore-savestate! savestate)
      (var ?frame-limit ?frame-limit)
      (var joypad (util.table-update savestate.joypad))
      (var joypad-schedule savestate.joypad-schedule)
      (let [
            ;; event-ids are the caller's desired events. The callbacks for each of
            ;; these call the result closure with their respective event-index.
            event-ids (icollect [event-index register (ipairs [...])]
                        (assert-event-id (register #(result event-index))))
            ;; limit-id is responsible for enforcing ?frame-limit. It gets called
            ;; at the end of every frame. If ?frame-limit is exceeded, it calls the
            ;; result closure with a nil argument.
            limit-id (assert-event-id
                       (bizhawk.event.onframeend
                         (lambda []
                           (set ?frame-limit (case ?frame-limit
                                               frame-limit (- frame-limit 1)))
                           (if (and ?frame-limit (<= ?frame-limit 0))
                             (result nil)))))
            ;; joypad-id also gets called at the end of every frame. Its purpose is
            ;; to advance the joyypad-schedule and to refresh the emulator's joypad
            ;; state for the new frame. The joypad-id callback never calls the
            ;; result closure.
            joypad-id (assert-event-id
                        (bizhawk.event.onframeend
                          (lambda []
                            (let [(buttons new-joypad-schedule) (joypad-schedule-advance joypad-schedule)]
                              (collect [k v (pairs buttons) &into joypad] (values k v))
                              (set joypad-schedule new-joypad-schedule)
                              (bizhawk.joypad.set joypad)))))
            ;; Now we yield, and wait for one of the event-ids or limit-id event
            ;; listeners to call the result closure and wake us back up.
            ?event-index (coroutine.yield)]
        ;; We're back. Unregister the event-ids and limit-id event listeners.
        (each [_ id (ipairs event-ids)]
          (assert (bizhawk.event.unregisterbyid id)))
        (assert (bizhawk.event.unregisterbyid limit-id))
        ;; We have to frame-align the emulator state before we can call
        ;; new-savestate! below. We keep the joypad-id listener registered until
        ;; after calling frame-align!, because frame-align! may cause the emulator
        ;; to advance by a frame. If that happens, we need to advance the
        ;; joypad-schedule and refresh the joypad inputs.
        (frame-align!)
        (assert (bizhawk.event.unregisterbyid joypad-id))
        ;; If it was one of the caller's events that happened (?event-index is not
        ;; nil), return a new savestate along with the index. If ?frame-limit was
        ;; exceeded, return nil.
        (case ?event-index
          event-index (values (util.table-update (new-savestate!)
                                                 {: joypad
                                                  : joypad-schedule})
                              event-index)
          _ nil)))))

;; Return a new savestate representing the state of the emulator after running
;; ?n frames. ?n defaults to 1.
(lambda next-frame [savestate ?n]
  (faccumulate [s savestate _ 1 (or ?n 1)]
    (assert (await-event s 1 bizhawk.event.onframeend))))

;; Return a new savestate representing the state of the emulator at the frame
;; boundary after the address addr is written, or nil if ?frame-limit is
;; reached.
(lambda await-write [savestate ?frame-limit addr]
  (await-event savestate ?frame-limit
               #(bizhawk.event.on_bus_write $ addr)))

;; Return a new savestate representing the state of the emulator at the frame
;; boundary after the address addr is executed, or nil if ?frame-limit is
;; reached.
(lambda await-exec [savestate ?frame-limit addr]
  (await-event savestate ?frame-limit
               #(bizhawk.event.on_bus_exec $ addr)))

;; Return a new savestate representing the frame boundary after the address
;; addr contains a u8 value that satisfies the ok? predicate, or nil if
;; ?frame-limit is reached.
(lambda await-u8-satisfies [savestate ?frame-limit addr ok?]
  (if
    (and ?frame-limit (<= ?frame-limit 0)) nil
    (ok? (read-u8 savestate addr)) savestate
    (await-event savestate ?frame-limit
                 (lambda [callback]
                   (bizhawk.event.on_bus_write
                     (lambda [addr val flags] (if (ok? val) (callback)))
                     addr)))))

;; Return a new savestate representing the frame boundary after the u8 at
;; address addr takes on a new value, or nil if ?frame-limit is reached.
(lambda await-u8-changes [savestate ?frame-limit addr]
  (let [current (read-u8 savestate addr)]
    (await-u8-satisfies savestate ?frame-limit addr #(not= $ current))))

;; Return a new savestate representing the first frame at which a savestate
;; satisfies the ok? predicate, or nil if ?frame-limit is reached.
(lambda await-frame-satisfies [savestate ?frame-limit ok?]
  (if
    (and ?frame-limit (<= ?frame-limit 0)) nil
    (ok? savestate) savestate
    (tail! (await-frame-satisfies
             (next-frame savestate)
             (if ?frame-limit (- ?frame-limit 1))
             ok?))))

;; Find the earliest frame at which running an action function on a savestate
;; produces a savestate that satisfies a test predicate.
;;
;; The action function takes a savestate as input. The output of the action
;; function is then given to the ok? function. If the ok? function returns a
;; result of nil or false, then this function returns three values: the
;; savestate before running the action, the result of running action, and the
;; result of running ok?. If ok? returns nil or false, then this function
;; advances the original savestate by one frame and tries again, continuing to
;; advance until ok? is finally satisfied.
;;
;; It is common to call this function with action and ok? functions that return
;; savestates (or nil). In that case, you can regard the return values as
;; "before action", "after action", and "after test".
(lambda earliest-effective-frame [savestate action ok?]
  (let [after-action (action savestate)
        after-test (ok? after-action)]
    (if after-test
      (values savestate after-action after-test)
      ;; No luck with this frame. Try again with the next frame.
      (tail! (earliest-effective-frame (next-frame savestate) action ok?)))))

;; Create a new savestate at the current emulator state, and pass it as an
;; argument to the given f function. The call happens inside a coroutine, to
;; insulate it from BizHawk's default auto-resume for scripts and make the
;; savestate manipulation of this module work. When f returns (with or without
;; an error), the emulator is paused.
(lambda run [f]
  (coroutine.resume
    (coroutine.create
      #(let [(ok err) (pcall f (joypad-release-all (new-savestate!)))]
         (bizhawk.client.pause)
         (collectgarbage) ; Encourage deletion of temporary savestate files.
         (if (not ok) (error err))))))

{: new-savestate!
 : restore-savestate!
 : joypad-set
 : joypad-press-release
 : joypad-release-all
 : frame-count
 : read-u8
 : next-frame
 : await-write
 : await-exec
 : await-u8-satisfies
 : await-u8-changes
 : await-frame-satisfies
 : earliest-effective-frame
 : run}
